/*****************************************\
*                                         *
* PCX.C - A class to combine several      *
*         small images in one pcx file    *
*         as well as read them out        *
*                                         *
* Copyright (C) 2000 by Josef Drexler     *
*               <jdrexler@julian.uwo.ca>  *
*                                         *
* Permission granted to copy and redist-  *
* ribute under the terms of the GNU GPL.  *
* For more info please read the file      *
* COPYING which should have come with     *
* this file.                              *
*                                         *
\*****************************************/

#include <stdlib.h>
#include <mem.h>

#include "pcx.h"
#include "error.h"

/***********************\
*			*
* class singlefile	*
*			*
\***********************/

singlefile::singlefile(char *filename, char *mode)
{
  FILE *f = fopen(filename, mode);
  if (!f) {
	fperror("\nCan't read %s", filename);
	exit(1);
  }
  setfile(f);
  thefilename = strdup(filename);
  autoclose = 1;
};

singlefile::singlefile(FILE *file)
{
  if (!file) {
	fperror("\nError opening file");
	exit(1);
  }
  setfile(file);
};

singlefile::~singlefile()
{
  if (autoclose && thefile)
	fclose(thefile);
  if (thefilename)
	free(thefilename);
};

void singlefile::setfile(FILE *file)
{
  thefile = file;
  thefilename = NULL;
  autoclose = 0;
}


/***********************\
*			*
* class pcxfile		*
*			*
\***********************/

pcxfile::pcxfile()
{
  if (sizeof(pcxheader) != 128) {
	printf("Somebody's packing the PCX header, it's not 128 bytes large!");
	exit(254);
  }

  sx = -1;
  sy = -1;
  totaly = 0;
  subx = 0;
  px = 0;
  py = 0;
  bandy = -1;
  bandlines = 0;
  band = NULL;
  run = 0;
  curfile = NULL;
  mfile = NULL;
  header.window[2] = -1;
  codecing = 0;
  notify = NULL;

  int i;
  for (i=0; i<256; i++)
	getcolourmap[i] = i;
  for (i=0; i<256; i++)
	putcolourmap[i] = i;
}

pcxfile::~pcxfile()
{
  if (mfile)
	delete(mfile);

  if (band) {
	for (int i=0; i<bandlines; i++)
		if (band[i])
			delete(band[i]);
	delete(band);
  }
}

void pcxfile::setfile(multifile *mfile)
{
  pcxfile::mfile = mfile;
  curfile = mfile->curfile();
}

void pcxfile::newfile(int sx)
{
  if (curfile) {    	// do we already have a file
	filedone(1);
  }

			// is it really a new file, or still the old one?
  if (getnextfile() || (header.window[2] == -1)) {
	curfile = getcurfile();
	newheader(sx);
	totaly = 0;
	subx = 0;
	px = 0;
	py = 0;
  }

  filestart();
  fseek(curfile, sizeof(header), SEEK_SET);
}

void pcxfile::newheader(int sx)
{
  const static pcxheader baseheader =
	{ 10, 5, 1, 8,		// manuf, version, encoding, bpp
	  {0, 0, -1, -1},	// window[4]
	  {72, 72},		// dpi[2]
	  {0},			// cmap[48]
	  0, 1, -1, 1,		// reserved, nplanes, bpl, palinfo
	  {-1, -1}		// screen[2]
	};			// filler[54]

  header = baseheader;
  header.window[2] = sx - 1;
  header.screen[0] = sx - 1;
  header.bpl = sx;
}

void pcxfile::startimage(int sx, int sy, int bandx, int bandy, bandnotify *notify)
{
  pcxfile::notify = notify;

  pcxfile::sx = sx;
  pcxfile::sy = sy;
  pcxfile::bandx = bandx;
  pcxfile::bandy = bandy;

  newfile(sx);

  if (bandy > bandlines)
	alloclines(bandy);

  thisbandy = bandy;
}

void pcxfile::alloclines(int newlines)
{
  U8 **newband, **oldband = band;

  if (newlines <= bandlines)
	return;

  newband = new U8*[newlines];
  if (!newband) {
	printf("Error allocating band array\n");
	exit(2);
  }

  int i;
  for (i=0; i<bandlines; i++)
	newband[i] = oldband[i];

  for (i=bandlines; i<newlines; i++) {
	newband[i] = new U8[sx];
	if (!(newband[i])) {
		printf("Error allocating new band lines\n");
		exit(2);
	}
  }

  if (band)
	delete(band);

  band = newband;
  i = bandlines;
  bandlines = newlines;

  for (; i<newlines; i++)
	initline(i);
}

void pcxfile::expirelines(int oldlines)
{
  for (int i=0; i<oldlines; i++) {
	U8 *old = band[0];
	memmove( &(band[0]), &(band[1]), (bandlines-1)*sizeof(band[0]));
	band[bandlines-1] = old;
	initline(bandlines-1);
  }
}

void pcxfile::initline(int y)
{
  if (y >= bandlines) {
	printf("y=%d larger than bandlines=%d\n", y, bandlines);
	exit(2);
  }
  setline(band[y]);
};


void pcxfile::endimage()
{
  if (px)
	newband();

  if (run)
	endencoding();

  if (band) {
	int i;
	for (i=0; i<bandlines; i++)
		if (band[i])
			delete(band[i]);
	delete(band);
	band = NULL;
  }

  notify = NULL;

  filedone(1);
}

void pcxfile::newline()
{
  cx = 0;
  cy++;
}

void pcxfile::streamputpixel(U8 colour)
{
  int x = subofsx(cx);
  int y = subofsy(cy);

  if (putcolourmap[colour] == -1) {
	printf("Agh! Putting colour %d but it has no map!\n", colour);
	exit(2);
  }

  band[y][x] = putcolourmap[colour];
  cx++;
}

void pcxfile::streamputpixel(U8 *buffer, unsigned long datasize)
{
  for (unsigned long i=0; i<datasize; i++) {
	streamputpixel(buffer[i]);
	if ( (i % (px-1)) == px-2)
		newline();
  }
}

U8 pcxfile::streamgetpixel()
{
  int x = subofsx(cx);
  int y = subofsy(cy);
  cx++;

  int colour = band[y][x];
  if (getcolourmap[colour] == -1) {
	printf("Agh! Getting colour %d but it has no map!\n", colour);
	exit(2);
  }

  return getcolourmap[colour];
}

void pcxfile::streamgetpixel(U8 *buffer, unsigned long datasize)
{
  for (unsigned long i=0; i<datasize; i++) {
	buffer[i] = streamgetpixel();
	if ( (i % px) == px-1)
		newline();
  }
}

void pcxfile::putpixel(int x, int y, U8 colour)
{
  if (putcolourmap[colour] == -1) {
	printf("Agh! Putting colour %d but it has no map!\n", colour);
	exit(2);
  }

  band[subofsy(y)][subofsx(x)] = putcolourmap[colour];
}

U8 pcxfile::getpixel(int x, int y)
{
  int colour = band[subofsy(y)][subofsx(x)];
  if (getcolourmap[colour] == -1) {
	printf("Agh! Getting colour %d but it has no map!\n", colour);
	exit(2);
  }

  return getcolourmap[colour];
}

void pcxfile::endsubimage()
{
}

void pcxfile::newband()
{
  int i, x, y;

  if ( (totaly + thisbandy > sy) && (thisbandy <= sy) ) {	// would be too large
	newfile(sx);
  }

  if (notify)
	notify->newband(this);

  totaly += thisbandy;

  for (y=0; y<thisbandy; y++) {
	startencoding();
	for (x=0; x<sx; x++)
		encodebyte(band[y][x]);
	endencoding();
  }

  for (i=0; i<bandlines; i++)
	initline(i);

  subx=0;
  px=0;

  thisbandy = bandy;

  // while testing only...
//  filedone(0);
}

void pcxfile::startencoding()
{
  if (codecing == 1)
	return;

  run = 0;
  codecing = 1;
}

void pcxfile::startdecoding()
{
  if (codecing == 2)
	return;

  run = 0;
  codecing = 2;
}

void pcxfile::encodebyte(U8 byte)
{
  if (codecing != 1) {
	printf("I'm not encoding, but got a byte?\n");
	exit(2);
  }

  if (run == 0)
	last = byte;

  if ( (last == byte) && (run < 0x3f) )
	run++;
  else {
	fflush(stdout);
	if (run >= 2) {
		fputc(run | 0xc0, curfile );
		fputc(last, curfile);
	} else {
		if ( (last & 0xc0) == 0xc0) {
			fputc(0xc1, curfile);
		}
		fputc(last, curfile);
	}
	last = byte;
	run = 1;
  }
}

U8 pcxfile::decodebyte()
{
  if (codecing != 2) {
	printf("I'm not decoding, but am supposed to return a byte?\n");
	exit(2);
  }

  if (run == 0) {
	last = fgetc(curfile);
	if ( (last & 0xc0) == 0xc0) {	// it's a run of equal bytes
		run = last & 0x3f;
		last = fgetc(curfile);
	} else
		run = 1;
  }

  run--;
  return last;
}

void pcxfile::endencoding()
{
  if (codecing == 1)
	encodebyte(last - 1);
  run = 0;
  codecing = 0;
}

int pcxfile::subofsx(int x)
{
  int ofsx = subx + x + dx;
  if (ofsx >= sx) {
	printf("\nofsx too large: is %d=%d+%d+%d, sx=%d\n", ofsx, subx, x, dx, sx);
	exit(2);
  }
  return ofsx;
}

int pcxfile::subofsy(int y)
{
  int ofsy =        y + dy;
  if (ofsy >= bandlines) {
	printf("\nofsy too large\n");
	exit(2);
  }
  return ofsy;
}

void pcxfile::installreadmap(int *map)
{
  for (int i=0; i<256; i++)
	getcolourmap[i] = map[i];
}

void pcxfile::installwritemap(int *map)
{
  for (int i=0; i<256; i++)
	putcolourmap[i] = map[i];
}
